#!/usr/bin/env python
# coding: utf-8

# In[1]:


useIlwis = False # False: use GDAL; True: use ILWIS


# In[2]:


import os
import glob
from zipfile import ZipFile
from netCDF4 import Dataset
from osgeo import gdal, osr
from osgeo.gdal import gdalconst
if useIlwis:
    import ilwis
import numpy as np
import pandas as pd
import numpy.ma as ma


# In[3]:


olci_files = 'olci'
climatology_files = 'data_in'
working_folder = 'data_out'


# In[4]:


def writeIlwis(data, filename, templateDataset):
    xsize = np.shape(data)[1]
    ysize = np.shape(data)[0]
    gdt = None
    if data.dtype == np.uint8:
        gdt = gdal.GDT_Byte
    else:
        gdt = gdal.GDT_Float32
    driver = gdal.GetDriverByName('ILWIS')
    dataset = driver.Create(filename, xsize, ysize, 1, gdt)
    if templateDataset is not None:
        dataset.SetGeoTransform(templateDataset.GetGeoTransform())
        dataset.SetProjection(templateDataset.GetProjection())
    dataset.GetRasterBand(1).WriteArray(data)
    dataset.FlushCache()
    dataset=None


# In[5]:


def preprocessWithIlwis(data_ndvi, data_nobs, date):
    # use GDAL to write the netCDF bands as ungeoreferenced ILWIS files
    # TODO: moving the data to ILWIS should not require storing and then reading a file; ILWIS should read the band from memory.
    writeIlwis(data_ndvi, working_folder + '/Afr_NDVI_' + date + '.mpr', None)
    writeIlwis(data_nobs, working_folder + '/Afr_NOBS_' + date + '.mpr', None)
    ilwis.Engine.setWorkingCatalog(ilwis_folder) # re-scan the catalog, as the writeIlwis() calls created files outside ilwis

    # now use ILWIS to set the georeference and rescale the raster images
    georef_afr = ilwis.GeoReference(ilwis_climatology + '/ndv_afr_olci300m_pc.grf')
    georef_eth = ilwis.GeoReference(ilwis_climatology + '/ndv_eth_olci300m_pc.grf')
    georef_eth_1km = ilwis.GeoReference(ilwis_climatology + '/ndv_eth_olci1km_pc.grf')

    # set the georeference (300m)
    raster_afr_ndvi = ilwis.RasterCoverage(ilwis_folder + '/Afr_NDVI_' + date + '.mpr')
    raster_afr_ndvi.setGeoReference(georef_afr)
    raster_afr_nobs = ilwis.RasterCoverage(ilwis_folder + '/Afr_NOBS_' + date + '.mpr')
    raster_afr_nobs.setGeoReference(georef_afr)

    # resample (pixel-perfect; this is creating a submap)
    raster_eth_ndvi = ilwis.Engine.do('resample', raster_afr_ndvi, georef_eth, 'nearestneighbour') # allowed values for the 3rd parameter are "nearestneighbour", "bilinear" and "bicubic"
    raster_eth_nobs = ilwis.Engine.do('resample', raster_afr_nobs, georef_eth, 'nearestneighbour')

    # scale to byte, for timesat
    raster_eth_ndvi = ilwis.Engine.do('mapcalc','iff((@1 > -0.08) and (@1 < 254), @1, -0.08)', raster_eth_ndvi)
    raster_eth_ndvi = ilwis.Engine.do('mapcalc','((@1 / 0.004) + (0.08 / 0.004))', raster_eth_ndvi)

    #integrate NOBS more than 0 and NDVI
    raster_eth_nobs = ilwis.Engine.do('mapcalc','iff((@1 > 0) and (@1 < 255), 1, 0)', raster_eth_nobs) # NOBS more than 0 observations will be assigned 1
    raster_eth_ndvi = ilwis.Engine.do('mapcalc','iff(@2 > 0, @1, ?)', raster_eth_ndvi, raster_eth_nobs)

    # aggregate 300m to 1km, by taking average of 3x3
    raster_eth_ndvi = ilwis.Engine.do('aggregateraster', raster_eth_ndvi, 'Avg', '3', 'yes')

    # overwrite the georeference with the official one
    raster_eth_ndvi.setGeoReference(georef_eth_1km)

    # store the result in the format required for timesat
    ilwis.Engine.do('setvaluerange', raster_eth_ndvi, 0, 255, 1) # TODO: does not work
    raster_eth_ndvi.store(ilwis_folder + '/timesat_input_' + date + '.mpr')
    
    # delete intermediate files
    os.unlink(working_folder + '/Afr_NDVI_' + date + '.mpr')
    os.unlink(working_folder + '/Afr_NDVI_' + date + '.mp#')
    os.unlink(working_folder + '/Afr_NOBS_' + date + '.mpr')
    os.unlink(working_folder + '/Afr_NOBS_' + date + '.mp#')


# In[6]:


def postProcessWithIlwis(file):
    # Zone adjust
    date = file.split('_')[-1][0:8]
    dekad = date[4:8]
    dekadfile = 'zone_adjust_' + dekad + '.mpr'
    raster_eth_ndvi = ilwis.RasterCoverage(ilwis_folder + '/' + file)
    zone_adjust = ilwis.RasterCoverage(ilwis_climatology + '/' + dekadfile)
    raster_eth_ndvi = ilwis.Engine.do('mapcalc','iff(@1 > 0, @1 - @2, @1)', raster_eth_ndvi, zone_adjust)
    ilwis.Engine.do('setvaluerange', raster_eth_ndvi, -255, 255, 1)
    raster_eth_ndvi.store(ilwis_folder + '/Eth_adjust_' + date + '.mpr')


# In[7]:


def preProcessWithGdal(data_ndvi, data_nobs, date):
    # geotransform = [minx, pixelsizex, 0, maxy, 0, -pixelsizey]
    geotransform = [-3339750.378000000500, 331.30800830026458333333333333333, 0, 4452945.286000000300, 0, -331.30800829613099702380952380952]
    platecarree = osr.SpatialReference()
    #platecarree.ImportFromEPSG(32662) # wrong?
    platecarree.ImportFromEPSG(4978)
    driver = gdal.GetDriverByName('MEM')
    xsize = np.shape(data_ndvi)[1]
    ysize = np.shape(data_ndvi)[0]
    raster_afr_ndvi = driver.Create('', xsize, ysize, 1, gdal.GDT_Float32)
    raster_afr_nobs = driver.Create('', xsize, ysize, 1, gdal.GDT_Byte)

    # set the georeference (300m)
    raster_afr_ndvi.SetGeoTransform(geotransform)
    raster_afr_ndvi.SetProjection(platecarree.ExportToWkt())
    raster_afr_nobs.SetGeoTransform(geotransform)
    raster_afr_nobs.SetProjection(platecarree.ExportToWkt())
    raster_afr_ndvi.GetRasterBand(1).WriteArray(data_ndvi)
    raster_afr_nobs.GetRasterBand(1).WriteArray(data_nobs)
    raster_afr_ndvi.FlushCache()
    raster_afr_nobs.FlushCache()

    # submap for Ethiopia
    raster_afr_ndvi = gdal.Translate('', raster_afr_ndvi, format='MEM', projWin = [3650183.186015000600, 1693150.423985000200, 5354762.687468000700, 356322.768326000080]) # ulx uly lrx lry
    raster_afr_nobs = gdal.Translate('', raster_afr_nobs, format='MEM', projWin = [3650183.186015000600, 1693150.423985000200, 5354762.687468000700, 356322.768326000080])

    # scale to byte, for timesat
    data_ndvi = raster_afr_ndvi.GetRasterBand(1).ReadAsArray()
    data_ndvi = np.where((data_ndvi > -0.08) & (data_ndvi < 254), data_ndvi, -0.08)
    data_ndvi = (data_ndvi / 0.004) + (0.08 / 0.004)

    #integrate NOBS more than 0 and NDVI
    data_nobs = raster_afr_nobs.GetRasterBand(1).ReadAsArray()
    data_nobs = np.where((data_nobs > 0) & (data_nobs < 255), 1, 0)
    data_ndvi = np.where(data_nobs > 0, data_ndvi, np.nan)

    # aggregate 300m to 1km, by taking average of 3x3 (imperfect)
    raster_afr_ndvi.GetRasterBand(1).WriteArray(data_ndvi) # does this work? or is raster_afr_ndvi read-only?
    raster_afr_ndvi.FlushCache()
    raster_afr_ndvi = gdal.Warp('', raster_afr_ndvi, format='MEM', resampleAlg = gdalconst.GRA_Average, outputType=raster_afr_ndvi.GetRasterBand(1).DataType, width=1715, height=1345)

    # store the result in the format required for timesat
    raster_afr_ndvi = gdal.Translate(working_folder + '/timesat_input_' + date + '.mpr', raster_afr_ndvi, format='ILWIS', outputType=gdal.GDT_Byte) # rounding?


# In[8]:


def postProcessWithGdal(file):
    # Zone adjust
    date = file.split('_')[-1][0:8]
    dekad = date[4:8]
    dekadfile = 'zone_adjust_' + dekad + '.mpr'
    raster_eth_ndvi = gdal.Open(working_folder + '/' + file)
    zone_adjust = gdal.Open(climatology_files + '/' + dekadfile)
    data_ndvi = raster_eth_ndvi.GetRasterBand(1).ReadAsArray()
    data_zone_adjust = zone_adjust.GetRasterBand(1).ReadAsArray()
    data_adjust = np.where(data_ndvi > 0, data_ndvi - data_zone_adjust, data_ndvi)
    writeIlwis(data_adjust, working_folder + '/Eth_adjust_' + date + '.mpr', raster_eth_ndvi)


# In[9]:


def executeTimesatWithIlwis():
    return


# In[10]:


def executeTimesatStub():
    timesatfiles = glob.glob(working_folder + '/timesat_input_*.mpr')
    for infile in timesatfiles:
        outfile = infile.replace('input','output')
        gdal.Translate(outfile, infile, format='ILWIS')


# In[11]:


def produceCSVFile():
    vici_grid = gdal.Open(climatology_files + '/vici_gridID.mpr')
    xsize = vici_grid.RasterXSize
    ysize = vici_grid.RasterYSize
    vici_grid = vici_grid.GetRasterBand(1).ReadAsArray()
    maxgridid = vici_grid.max()
    adjusted_files = glob.glob(working_folder + '/Eth_adjust_*.mpr')
    adjusted_files.sort()
    adjusted_files = adjusted_files[-5:]
    csv = np.empty((maxgridid,len(adjusted_files)+1),dtype='uint32')
    for gridid in range(maxgridid):
        csv[gridid,0] = gridid+1
    for ix, file in enumerate(adjusted_files):
        file = gdal.Open(file)
        file = file.GetRasterBand(1).ReadAsArray()
        for x in range(xsize):
            for y in range(ysize):
                gridid = vici_grid[y,x]
                if gridid > 0:
                    value = file[y,x]
                    csv[gridid-1,ix+1] = value
    column_names = ['VICI_GridID']
    for file in adjusted_files:
        file = file[:-4] # remove extension
        file = file.split('_')
        date = file[-1] # last component is the date
        column_names.append('ndvi_' + date)
    csv = pd.DataFrame(csv)
    csv.columns = column_names
    # outputfilename is related to last date
    outputfile = column_names[-1].replace('ndvi_', 'VICI_') + '.csv'
    csv.to_csv(working_folder + '/' + outputfile)
    return csv # tempoary return this, so we can print it, for confirmation that the content is right


# In[12]:


if useIlwis:
    # ilwis needs to scan the catalogs
    # os.getcwd() is used because ilwis requires  absolute paths
    ilwis.disconnectIssueLogger() # make silent (not working)
    ilwis_climatology = os.getcwd() + '/' + climatology_files
    ilwis.Engine.setWorkingCatalog(ilwis_climatology)
    ilwis_folder = os.getcwd() + '/' + working_folder
    ilwis.Engine.setWorkingCatalog(ilwis_folder)

print('Preprocessing: OLCI submap, scale, filter, aggregate.')
olcifiles = glob.glob(olci_files + '/c_gls_NDVI300_*.zip')
olcifiles = olcifiles[-36:] # incorporate climatology; the following sequence has to be repeated 5 times, to reprocess the most recent (previously processed) data

seq = 1
for file in olcifiles:
    file = file.replace('\\','/') # in case of Windows host
    zipname = file.split('/')[-1]
    date = zipname.split('_')[3][0:8]
    print('[' + str(seq) + '/' + str(len(olcifiles)) + '] Processing: ' + date)
    ncfile = zipname[:-4] + '.nc'
    z = ZipFile(file)
    f = z.open(date + '/' + ncfile)
    f = Dataset('in-mem-file', mode='r', memory=f.read()) # FileNotFoundError in netCDF < 1.5; netCDF minimum version 1.5 required
    data_ndvi = f.variables['NDVI'][0,:,:]
    mask_ndvi = ma.getmaskarray(data_ndvi)
    data_nobs = f.variables['NOBS'][0,:,:]
    mask_nobs = ma.getmaskarray(data_nobs)
    if useIlwis:
        preprocessWithIlwis(data_ndvi, data_nobs, date)
    else:
        preProcessWithGdal(data_ndvi, data_nobs, date)
    f.close()
    seq = seq + 1

if useIlwis:
    ilwis.connectIssueLogger() # re-enable messages


# In[13]:


print('Running timesat')
executeTimesatStub()


# In[14]:


print('Post processing: zone-adjust.')
filteredfiles = glob.glob(working_folder + '/timesat_output*.mpr')
for file in filteredfiles:
    file = file.replace('\\','/') # in case of Windows host
    file = file.split('/')[-1]
    if useIlwis:
        postProcessWithIlwis(file)
    else:
        postProcessWithGdal(file)


# In[15]:


print('Producing CSV file')
csv = produceCSVFile()
print(csv.head())
print('...')
print(csv.tail())


# In[16]:


print('Cleanup: delete intermediate data files')
timesatfiles = glob.glob(working_folder + '/timesat_*.*') # both input and output
for file in timesatfiles:
    os.unlink(file)


# In[ ]:





# In[ ]:




